/*
oop_ex85.cpp
overloading unary operator, as non function member
*/


#include<cstdio>
#include<iostream>
#include<cassert>
#include<cmath>

using namespace std;

class Matrix
{
	int row;
	int col;
	double** element;

public:

	Matrix(int r = 0, int c = 0);
	Matrix(const Matrix& X);
	~Matrix();

	double& operator()(int i, int j);

	//---binary operators, as function members---
	Matrix& operator=(Matrix& X);
	Matrix& operator+(Matrix& X);
	Matrix& operator-(Matrix& X);
	Matrix& operator*(Matrix& X);
	//-------------------------------------------


	//---unary operators, as non function members---
	friend Matrix& operator-(Matrix&);
	friend bool operator!(Matrix&);		//isEmpty?
										//-------------------------------------------

	void print();
};

//friend functions
//===================================================
Matrix& operator-(Matrix& m)
{
	for (int i = 0; i<m.row; i++)
	{
		for (int j = 0; j<m.col; j++)
		{
			m.element[i][j] *= (-1);
		}
	}
	return m;
}

bool operator!(Matrix& m)
{
	if (m.row == 0 || m.col == 0)
		return true;	//matrix is empty
	else
		return false;
}

// function member definitions
//===================================================
Matrix::Matrix(int r, int c) :row(r), col(c)
{
	element = new double*[row];

	for (int i = 0; i<row; i++)
		element[i] = new double[col];

	for (int i = 0; i<row; i++)
		for (int j = 0; j<col; j++)
			element[i][j] = 0.0;
}


Matrix::Matrix(const Matrix& X)
{
	row = X.row;
	col = X.col;

	element = new double*[row];

	for (int i = 0; i<row; i++)
		element[i] = new double[col];

	for (int i = 0; i<row; i++)
		for (int j = 0; j<col; j++)
			element[i][j] = X.element[i][j];
}


Matrix::~Matrix()
{
	for (int i = 0; i<row; i++)
		delete[] element[i];

	delete[] element;
}


double& Matrix::operator()(int i, int j)
{
	return element[i][j];
}



Matrix& Matrix::operator=(Matrix& X)
{
	if (this != &X)
	{
		// release the original matrix using delete
		for (int i = 0; i<row; i++)
			delete[] element[i];

		delete[] element;

		// create the matrix using new
		row = X.row;
		col = X.col;

		element = new double*[row];

		for (int i = 0; i<row; i++)
			element[i] = new double[col];

		for (int i = 0; i<row; i++)
			for (int j = 0; j<col; j++)
				element[i][j] = X.element[i][j];
	}

	return *this;
}


Matrix& Matrix::operator+(Matrix& X)
{
	assert(row == X.row && col == X.col);

	Matrix* ans = new Matrix(row, col);

	for (int i = 0; i<row; i++)
	{
		for (int j = 0; j<col; j++)
		{
			ans->element[i][j] = element[i][j] + X.element[i][j];
		}
	}

	return(*ans);
}


Matrix& Matrix::operator-(Matrix& X)
{
	assert(row == X.row && col == X.col);

	Matrix* ans = new Matrix(row, col);

	for (int i = 0; i<row; i++)
	{
		for (int j = 0; j<col; j++)
		{
			ans->element[i][j] = element[i][j] - X.element[i][j];
		}
	}

	return(*ans);
}


Matrix& Matrix::operator*(Matrix& X)
{
	assert(col == X.row);

	Matrix* ans = new Matrix(row, X.col);

	for (int i = 0; i<row; i++)
	{
		for (int j = 0; j<X.col; j++)
		{
			for (int k = 0; k<col; k++)
				ans->element[i][j] += element[i][k] * X.element[k][j];
		}
	}

	return(*ans);
}


void Matrix::print()
{
	for (int i = 0; i<row; i++)
	{
		for (int j = 0; j<col; j++)
		{
			cout << element[i][j] << " ";
		}
		cout << endl;
	}
}


// main
//==================================================
void main(void)
{
	int r = 3, c = 3;
	int i, j;

	Matrix A(r, c), B(r, c), C;
	Matrix ans1(r, c), ans2(r, c), ans3(r, c);

	for (i = 0; i<r; i++)
	{
		for (j = 0; j<c; j++)
		{
			A(i, j) = i + j;
			B(i, j) = j - i;
		}
	}

	if (!A == false)
	{
		cout << "A = " << endl;
		A.print();
		cout << "-A = " << endl;
		(-A).print();
	}

	if (!B == false)
	{
		cout << "B = " << endl;
		B.print();
	}

	if (!C == false)
	{
		cout << "C = " << endl;
		C.print();
	}

	cout << endl;


	ans1 = -A + B;
	ans2 = -A - B;
	ans3 = -A * B - ans2;


	cout << "ans1 = " << endl;
	ans1.print();
	cout << "ans2 = " << endl;
	ans2.print();
	cout << "ans3 = " << endl;
	ans3.print();
	cout << endl;

	system("pause");
}

