/*
oop_ex86.cpp
overloading uniary and binary operator, as non function member
*/


#include<cstdio>
#include<iostream>
#include<cassert>
#include<cmath>

using namespace std;

class Matrix
{
	int row;
	int col;
	double** element;

public:

	Matrix(int r = 0, int c = 0);
	Matrix(const Matrix& X);
	~Matrix();

	//---operators which cannot be non function members---
	double& operator()(int i, int j) { return element[i][j]; }

	Matrix& operator=(Matrix& X)
	{
		if (this != &X)
		{
			// release the original matrix using delete
			for (int i = 0; i<row; i++)
				delete[] element[i];

			delete[] element;

			// create the matrix using new
			row = X.row;
			col = X.col;

			element = new double*[row];

			for (int i = 0; i<row; i++)
				element[i] = new double[col];

			for (int i = 0; i<row; i++)
				for (int j = 0; j<col; j++)
					element[i][j] = X.element[i][j];
		}

		return *this;
	}

	//friend double& operator()(Matrix& m, int i, int j);
	//friend Matrix& operator=(Matrix& l, Matrix& r);
	//-----------------------------------------------


	//---binary operators, as non function members---
	friend Matrix& operator+(Matrix& l, Matrix& r);
	friend Matrix& operator-(Matrix& l, Matrix& r);
	friend Matrix& operator*(Matrix& l, Matrix& r);
	//-----------------------------------------------


	//---unary operators, as non function members---
	friend Matrix& operator-(Matrix&);
	friend bool operator!(Matrix&);		//isEmpty?
										//-----------------------------------------------

	void print();
};

//friend functions
//===================================================
//double& operator()(Matrix& m, int i, int j){ return m.element[i][j]; }
/*
Matrix& operator=(Matrix& l, Matrix& r)
{
assert(l.row == r.row && l.col == r.col);
for(int i=0;i<l.row;i++)
for(int j=0;j<l.col;j++)
l.element[i][j] = r.element[i][j];

return l;
}
*/

// unary
Matrix& operator-(Matrix& m)
{
	for (int i = 0; i<m.row; i++)
	{
		for (int j = 0; j<m.col; j++)
		{
			m.element[i][j] *= (-1);
		}
	}
	return m;
}

bool operator!(Matrix& m)
{
	if (m.row == 0 || m.col == 0)
		return true;	//matrix is empty
	else
		return false;
}

//	binary
Matrix& operator+(Matrix& l, Matrix& r)
{
	assert(l.row == r.row && l.col == r.col);

	Matrix* ans = new Matrix(l.row, l.col);

	for (int i = 0; i<ans->row; i++)
	{
		for (int j = 0; j<ans->col; j++)
		{
			ans->element[i][j] = l.element[i][j] + r.element[i][j];
		}
	}

	return(*ans);
}


Matrix& operator-(Matrix& l, Matrix& r)
{
	assert(l.row == r.row && l.col == r.col);

	Matrix* ans = new Matrix(l.row, l.col);

	for (int i = 0; i<ans->row; i++)
	{
		for (int j = 0; j<ans->col; j++)
		{
			ans->element[i][j] = l.element[i][j] - r.element[i][j];
		}
	}

	return(*ans);
}


Matrix& operator*(Matrix& l, Matrix& r)
{
	assert(l.col == r.row);

	Matrix* ans = new Matrix(l.row, r.col);

	for (int i = 0; i<ans->row; i++)
	{
		for (int j = 0; j<ans->col; j++)
		{
			for (int k = 0; k<l.col; k++)
				ans->element[i][j] += l.element[i][k] * r.element[k][j];
		}
	}

	return(*ans);
}


// function member definitions
//===================================================
Matrix::Matrix(int r, int c) :row(r), col(c)
{
	element = new double*[row];

	for (int i = 0; i<row; i++)
		element[i] = new double[col];

	for (int i = 0; i<row; i++)
		for (int j = 0; j<col; j++)
			element[i][j] = 0.0;
}


Matrix::Matrix(const Matrix& X)
{
	row = X.row;
	col = X.col;

	element = new double*[row];

	for (int i = 0; i<row; i++)
		element[i] = new double[col];

	for (int i = 0; i<row; i++)
		for (int j = 0; j<col; j++)
			element[i][j] = X.element[i][j];
}


Matrix::~Matrix()
{
	for (int i = 0; i<row; i++)
		delete[] element[i];

	delete[] element;
}


void Matrix::print()
{
	for (int i = 0; i<row; i++)
	{
		for (int j = 0; j<col; j++)
		{
			cout << element[i][j] << " ";
		}
		cout << endl;
	}
}


// main
//==================================================
void main(void)
{
	int r = 3, c = 3;
	int i, j;

	Matrix A(r, c), B(r, c), C;
	Matrix ans1(r, c), ans2(r, c), ans3(r, c);

	for (i = 0; i<r; i++)
	{
		for (j = 0; j<c; j++)
		{
			A(i, j) = i + j;
			B(i, j) = j - i;
		}
	}

	if (!A == false)
	{
		cout << "A = " << endl;
		A.print();
		cout << "-A = " << endl;
		(-A).print();
	}

	if (!B == false)
	{
		cout << "B = " << endl;
		B.print();
	}

	if (!C == false)
	{
		cout << "C = " << endl;
		C.print();
	}

	cout << endl;


	ans1 = -A + B;
	ans2 = -A - B;
	ans3 = -A * B - ans2;


	cout << "ans1 = " << endl;
	ans1.print();
	cout << "ans2 = " << endl;
	ans2.print();
	cout << "ans3 = " << endl;
	ans3.print();
	cout << endl;

	system("pause");
}

