/*
oop_ex87.cpp
overloading operator as non function member
, which is not necessary to be 'friend'
*/


#include<cstdio>
#include<iostream>
#include<cassert>
#include<cmath>

using namespace std;

class Matrix
{
	int row;
	int col;
	double** element;

public:

	Matrix(int r = 0, int c = 0);
	Matrix(const Matrix& X);
	~Matrix();

	//---operators which cannot be non function members---
	double& operator()(int i, int j) { return element[i][j]; }

	Matrix& operator=(Matrix& X)
	{
		if (this != &X)
		{
			// release the original matrix using delete
			for (int i = 0; i<row; i++)
				delete[] element[i];

			delete[] element;

			// create the matrix using new
			row = X.row;
			col = X.col;

			element = new double*[row];

			for (int i = 0; i<row; i++)
				element[i] = new double[col];

			for (int i = 0; i<row; i++)
				for (int j = 0; j<col; j++)
					element[i][j] = X.element[i][j];
		}

		return *this;
	}
	//-----------------------------------------------


	void print();

	// need access functions for private members
	int getRow() { return row; }
	int getCol() { return col; }
};

//non function members
//===================================================

// unary
Matrix& operator-(Matrix& m)
{
	for (int i = 0; i<m.getRow(); i++)
	{
		for (int j = 0; j<m.getCol(); j++)
		{
			m(i, j) *= (-1);
		}
	}
	return m;
}

bool operator!(Matrix& m)
{
	if (m.getRow() == 0 || m.getCol() == 0)
		return true;	//matrix is empty
	else
		return false;
}

//	binary
Matrix& operator+(Matrix& l, Matrix& r)
{
	assert(l.getRow() == r.getRow() && l.getCol() == r.getCol());

	Matrix* ans = new Matrix(l.getRow(), l.getCol());

	for (int i = 0; i<ans->getRow(); i++)
	{
		for (int j = 0; j<ans->getCol(); j++)
		{
			(*ans)(i, j) = l(i, j) + r(i, j);
		}
	}

	return(*ans);
}


Matrix& operator-(Matrix& l, Matrix& r)
{
	assert(l.getRow() == r.getRow() && l.getCol() == r.getCol());

	Matrix* ans = new Matrix(l.getRow(), l.getCol());

	for (int i = 0; i<ans->getRow(); i++)
	{
		for (int j = 0; j<ans->getCol(); j++)
		{
			(*ans)(i, j) = l(i, j) - r(i, j);
		}
	}

	return(*ans);
}


Matrix& operator*(Matrix& l, Matrix& r)
{
	assert(l.getCol() == r.getRow());

	Matrix* ans = new Matrix(l.getRow(), r.getCol());

	for (int i = 0; i<ans->getRow(); i++)
	{
		for (int j = 0; j<ans->getCol(); j++)
		{
			for (int k = 0; k<l.getCol(); k++)
				(*ans)(i, j) += l(i, k) * r(k, j);
		}
	}

	return(*ans);
}


// function member definitions
//===================================================
Matrix::Matrix(int r, int c) :row(r), col(c)
{
	element = new double*[row];

	for (int i = 0; i<row; i++)
		element[i] = new double[col];

	for (int i = 0; i<row; i++)
		for (int j = 0; j<col; j++)
			element[i][j] = 0.0;
}


Matrix::Matrix(const Matrix& X)
{
	row = X.row;
	col = X.col;

	element = new double*[row];

	for (int i = 0; i<row; i++)
		element[i] = new double[col];

	for (int i = 0; i<row; i++)
		for (int j = 0; j<col; j++)
			element[i][j] = X.element[i][j];
}


Matrix::~Matrix()
{
	for (int i = 0; i<row; i++)
		delete[] element[i];

	delete[] element;
}



void Matrix::print()
{
	for (int i = 0; i<row; i++)
	{
		for (int j = 0; j<col; j++)
		{
			cout << element[i][j] << " ";
		}
		cout << endl;
	}
}


// main
//==================================================
void main(void)
{
	int r = 3, c = 3;
	int i, j;

	Matrix A(r, c), B(r, c), C;
	Matrix ans1(r, c), ans2(r, c), ans3(r, c);

	for (i = 0; i<r; i++)
	{
		for (j = 0; j<c; j++)
		{
			A(i, j) = i + j;
			B(i, j) = j - i;
		}
	}

	if (!A == false)
	{
		cout << "A = " << endl;
		A.print();
		cout << "-A = " << endl;
		(-A).print();
	}

	if (!B == false)
	{
		cout << "B = " << endl;
		B.print();
	}

	if (!C == false)
	{
		cout << "C = " << endl;
		C.print();
	}

	cout << endl;


	ans1 = -A + B;
	ans2 = -A - B;
	ans3 = -A * B - ans2;


	cout << "ans1 = " << endl;
	ans1.print();
	cout << "ans2 = " << endl;
	ans2.print();
	cout << "ans3 = " << endl;
	ans3.print();
	cout << endl;

	system("pause");
}

