/*
oop_ex88.cpp
overloading << and >> operators
*/


#include<cstdio>
#include<iostream>
#include<cassert>
#include<cmath>

using namespace std;

class Matrix
{
	int row;
	int col;
	double** element;

public:

	Matrix(int r = 0, int c = 0);
	Matrix(const Matrix& X);
	~Matrix();

	double& operator()(int i, int j);

	//---binary operators, as function members---
	Matrix& operator=(Matrix& X);
	Matrix& operator+(Matrix& X);
	Matrix& operator-(Matrix& X);
	Matrix& operator*(Matrix& X);
	//-------------------------------------------


	//---unary operators, as function members---
	Matrix& operator-();
	bool operator!();
	//-----------------------------------------


	//--- overloading of << and >> ---
	friend ostream& operator<<(ostream&, Matrix&);
	friend istream& operator>>(istream&, Matrix&);

	//friend void operator<<(ostream&, Matrix&);	//this cannot support
	//friend void operator>>(istream&, Matrix&);	//chained function calls

	//ostream& operator<<(ostream&){ cout<<"operator<<"<<endl; }	//error, can't be 
	//istream& operator>>(istream&){ cout<<"operator>>"<<endl; }	//function member of Matrix
	//-----------------------------------------


	void print();
};


ostream& operator<<(ostream& os, Matrix& m)
{
	for (int i = 0; i<m.row; i++)
	{
		for (int j = 0; j<m.col; j++)
		{
			os << m.element[i][j] << " ";
		}
		cout << endl;
	}

	return os;
}

istream& operator>>(istream& is, Matrix& m)
{
	is >> m.row;
	m.col = m.row;

	m.element = new double*[m.row];

	for (int i = 0; i<m.row; i++)
	{
		m.element[i] = new double[m.col];
	}

	for (int i = 0; i<m.row; i++)
	{
		for (int j = 0; j<m.col; j++)
			m.element[i][j] = 0.0;
	}

	return is;
}


/*
// can't support chained function call
void operator<<(ostream& os, Matrix& m)
{
for(int i=0;i<m.row;i++)
{
for(int j=0;j<m.col;j++)
{
os<<m.element[i][j]<<" ";
}
cout<<endl;
}
}
*/
/*
// can't support chained function call
void operator>>(istream& is, Matrix& m)
{
is >> m.row;
m.col = m.row;

m.element = new double*[m.row];

for(int i=0;i<m.row;i++)
{
m.element[i] = new double[m.col];
}

for(int i=0;i<m.row;i++)
{
for(int j=0;j<m.col;j++)
m.element[i][j]=0.0;
}
}
*/


//function member definition
//===================================================
Matrix::Matrix(int r, int c) :row(r), col(c)
{
	element = new double*[row];

	for (int i = 0; i<row; i++)
		element[i] = new double[col];

	for (int i = 0; i<row; i++)
		for (int j = 0; j<col; j++)
			element[i][j] = 0.0;
}


Matrix::Matrix(const Matrix& X)
{
	row = X.row;
	col = X.col;

	element = new double*[row];

	for (int i = 0; i<row; i++)
		element[i] = new double[col];

	for (int i = 0; i<row; i++)
		for (int j = 0; j<col; j++)
			element[i][j] = X.element[i][j];
}


Matrix::~Matrix()
{
	for (int i = 0; i<row; i++)
		delete[] element[i];

	delete[] element;
}


double& Matrix::operator()(int i, int j)
{
	return element[i][j];
}

//---unary operators
Matrix& Matrix::operator-()
{
	for (int i = 0; i<row; i++)
	{
		for (int j = 0; j<col; j++)
		{
			element[i][j] *= (-1);
		}
	}
	return *this;
}

bool Matrix::operator!()
{
	if (row == 0 || col == 0)
		return true;
	else
		return false;
}

//---binary operators
Matrix& Matrix::operator=(Matrix& X)
{
	if (this != &X)
	{
		// release the original matrix using delete
		for (int i = 0; i<row; i++)
			delete[] element[i];

		delete[] element;

		// create the matrix using new
		row = X.row;
		col = X.col;

		element = new double*[row];

		for (int i = 0; i<row; i++)
			element[i] = new double[col];

		for (int i = 0; i<row; i++)
			for (int j = 0; j<col; j++)
				element[i][j] = X.element[i][j];
	}

	return *this;
}


Matrix& Matrix::operator+(Matrix& X)
{
	assert(row == X.row && col == X.col);

	Matrix* ans = new Matrix(row, col);

	for (int i = 0; i<row; i++)
	{
		for (int j = 0; j<col; j++)
		{
			ans->element[i][j] = element[i][j] + X.element[i][j];
		}
	}

	return(*ans);
}


Matrix& Matrix::operator-(Matrix& X)
{
	assert(row == X.row && col == X.col);

	Matrix* ans = new Matrix(row, col);

	for (int i = 0; i<row; i++)
	{
		for (int j = 0; j<col; j++)
		{
			ans->element[i][j] = element[i][j] - X.element[i][j];
		}
	}

	return(*ans);
}


Matrix& Matrix::operator*(Matrix& X)
{
	assert(col == X.row);

	Matrix* ans = new Matrix(row, X.col);

	for (int i = 0; i<row; i++)
	{
		for (int j = 0; j<X.col; j++)
		{
			for (int k = 0; k<col; k++)
				ans->element[i][j] += element[i][k] * X.element[k][j];
		}
	}

	return(*ans);
}


void Matrix::print()
{
	for (int i = 0; i<row; i++)
	{
		for (int j = 0; j<col; j++)
		{
			cout << element[i][j] << " ";
		}
		cout << endl;
	}
}


//main
//==================================================
void main(void)
{
	int r = 3, c = 3;
	int i, j;

	//Matrix A(r,c), B(r,c), C;
	//Matrix ans1(r,c), ans2(r,c), ans3(r,c);
	Matrix A, B, C, ans1, ans2, ans3;

	cout << "Please enter the matrix size 7 times." << endl;
	cin >> r >> c >> A >> B >> ans1 >> ans2 >> ans3;

	for (i = 0; i<r; i++)
	{
		for (j = 0; j<c; j++)
		{
			A(i, j) = i + j;
			B(i, j) = j - i;
		}
	}

	if (!A == false)
	{
		cout << "A = " << endl;
		A.print();
		cout << "-A = " << endl;
		(-A).print();
	}

	if (!B == false)
	{
		cout << "B = " << endl;
		B.print();
	}

	if (!C == false)
	{
		cout << "C = " << endl;
		C.print();
	}

	cout << endl;


	ans1 = -A + B;
	ans2 = -A - B;
	ans3 = -A * B - ans2;


	cout << "ans1 = " << endl;
	//ans1.print();
	cout << ans1;
	cout << "ans2 = " << endl;
	//ans2.print();
	cout << ans2;
	cout << "ans3 = " << endl;
	//ans3.print();
	cout << ans3;
	cout << endl;

	//cout<<ans1<<ans2<<ans3;

	system("pause");
}



